import java.io.*;
import java.util.*;

public class SMLParser {
	private static final int DOCUMENT = 0;
	private static final int STATEMENT = 1;
	private static final int TAG = 2;
	private static final int TEXT = 3;
	
    public Statement parseGrammar( InputStream in ) {    	
    	Statement rc = null;
    	ArrayList parseStack = new ArrayList();
    	try {
    		push( parseStack, new ParseState( DOCUMENT ) );
        	rc = (Statement)readProduction( parseStack, in );
        	if( !(rc instanceof MultiStatement ) ) {
        		rc = new MultiStatement(rc);
        	}
    	} catch( ParseException e ) {
    		System.err.println(e);
    	} catch( IOException e ) {
    		System.err.println(e);
    	}
        return rc;
    }
    
    private Statement readProduction( ArrayList stack, InputStream in ) 
    	throws ParseException, IOException {
    	Statement rc = new EmptyStatement();
    	while( !stack.isEmpty() ) {
    		ParseState state = pop(stack);
    		switch( state.m_nProduction ) {
    			case DOCUMENT: {
    				// get or create the statement list
    				ArrayList statements = (ArrayList)state.getProperty("statementlist");
    				if( statements == null ) {
    					statements = new ArrayList();
    					state.setProperty("statementlist", statements);
    				}
    				if( state.state() == 0 ) {
    					// we are entering DOCUMENT for the first time
    					// look for a STATEMENT and keep LOOPING
    					state.incrementState();
    					push( stack, state );
    					push( stack, new ParseState( STATEMENT ) );
    				} else {
    					// we are looping in DOCUMENT
    					// rc should be the child we are looking for
    					if( rc != null ) {
    						// if there is a current Statement, add it to the list
    						statements.add(rc);
    						rc = null;
    						// keep parsing DOCUMENT and STATEMENT
    						push( stack, state );
    						push( stack, new ParseState( STATEMENT ) );
    					} else {
    						// there are no more statements - return a final value
    						if( statements.size() == 0 ) {
    							rc = new EmptyStatement();
    						} else if( statements.size() == 1 ) {
    							rc = (Statement)statements.get(0);
    						} else {
    							rc = new MultiStatement( statements );
    						}
    					}
    				}
    			} break;
    			case STATEMENT: {
    				if( state.state() == 0 ) {
    					// we are entering statement for the first time
    					// we need a look ahead here
    					in.mark(1);
    					// read a character
    					char c = 0;
    					try {
                			c = readChar(in);                    
    					} catch( ParseException e ) {
    						if( e.wrappedException() instanceof EOFException ) {
    							// if the first character does not exist, there is no statement
    							rc = null;
    						} else {
    							// another error occurred - throw it up the call stack
    							throw e;
    						}
    					}
						if( c == '<' ) {
							// parse a TAG
							in.mark(1);
							if( readChar(in) == '/' ) {
								// this is a closing tag
								// fall back and let the parent tag close
								// rc will be handled by parent
								in.reset();
							} else {
								in.reset();
								push(stack, new ParseState(TAG) );
							}
                		} else {
                			// parse TEXT
                    		in.reset();
							push(stack, new ParseState(TEXT) );
                		}
    				}
    				
    			} break;
    			case TAG: {
    				if( state.state() == 0 ) {
    					// Get the opening of the tag
    					state.incrementState();
                		StringBuffer tagLabel = new StringBuffer();
                		try {
                			for( char c = readChar(in); c != '>'; c = readChar(in) ) {
                				tagLabel.append(c);
                			}
            			} catch( ParseException e ) {
            				e.setMessage( "expected '>'" );
            				throw( e );
                		}
                		state.setProperty("label", tagLabel.toString());
                		// we want to come back to TAG 
                		push( stack, state );
                		// after we parse a DOCUMENT
                		push( stack, new ParseState(DOCUMENT) );
                		// set up a default Statement in case there is nothing in tag
                		rc = new EmptyStatement();
    				} else if( state.state() == 1 ) {
    					// we should have just parsed a DOCUMENT
    					String label = (String)state.getProperty( "label" );
    					rc = TagFactory.createTag(label, rc);
    					
    					// the the TagFactor cannot create a tag, it is not valid
    					if( rc == null ) {
    						ParseException e = new ParseException( null, lineCount() );
    						e.setMessage("tag id '"+label+"' is invalid.");
    						throw e;
    					}
    					
    					// make sure we have the right closing tag
    					StringBuffer closingTag = new StringBuffer();
                		try {
                			// '<' has already been eaten by STATEMENT
    						char c = readChar(in);
    						if( c != '/' ) {
    							ParseException e = new ParseException( null, lineCount() );
    							e.setMessage("expected '/'");
    							throw e;
    						}
                			for( c = readChar(in); c != '>'; c = readChar(in) ) {
                				closingTag.append(c);
                			}
            			} catch( ParseException e ) {
            				e.setMessage( "expected '>'" );
            				throw(e);
                		}
    					if( !label.equals(closingTag.toString()) ) {
    						ParseException e = new ParseException( null, lineCount() );
    						e.setMessage( "expected closing tag for '"+label+"'. Found '"+closingTag.toString()+"'." );
    						throw(e);
    					}    			
    				}
    			} break;
    			case TEXT: {
    				StringBuffer text = new StringBuffer();
    				try {
    					in.mark(1);
    					for( char c = readChar(in); c != '<'; c = readChar(in) ) {
    						if( c == '>' ) {
    							in.reset();
    							ParseException e = new ParseException( null, lineCount() );
    							e.setMessage( "expected '<'. Found '>'." );
    							throw(e);
    						}
    						text.append(c);
    						in.mark(1);
    					}
    					in.reset();
    				} catch( ParseException e ) {
    					if( e.wrappedException() instanceof EOFException ) {
    						// fall through and return String()
    					} else {
    						throw(e);
    					}
    				}
    				if( text.length() > 0 ) {
    					rc = new Text(text.toString());
    				}
    			} break;
    		}
    	}
    	return rc;
    }
    
    private char readChar( InputStream in )
    	throws ParseException {
    	int rc = 0;
    	try {
    		rc = in.read();
    		if( rc == '\n' ) {
    			m_nLineCount++;
    		} else if( rc == -1 ) {
    			throw new EOFException();
    		}
    	} catch( IOException e ) {
    		throw new ParseException(e, lineCount());
    	}
    	return (char)rc;
    }
    
    private int m_nLineCount = 0;
    private int lineCount() {
    	return m_nLineCount;
    }
    
    private static void push( ArrayList stack, ParseState obj ) {
    	stack.add(obj);
    }
    
    private static ParseState pop( ArrayList stack ) {
    	return (ParseState)stack.remove(stack.size()-1);
    }
}